use std::fmt::Display;
use std::str::FromStr;

use anyhow::{anyhow, Result};
use isahc::http::StatusCode;
use isahc::{ReadResponseExt, RequestExt};
use serde::{Deserialize, Serialize};

use crate::{deserialize_normal_response, sign, BunqClient, BASE_URL};

pub(super) fn get(client: &BunqClient) -> anyhow::Result<Vec<MonetaryAccount>> {
    let response = isahc::http::Request::get(format!(
        "{}/v1/user/{}/monetary-account",
        BASE_URL, client.user_id
    ))
    .header("X-Bunq-Client-Authentication", &client.token)
    .body(())?
    .send()?
    .text()?;

    let accounts = crate::deserialize_normal_response::<Vec<MonetaryAccount>>(&response)?;

    Ok(accounts.response)
}

pub(super) fn request_statement(
    client: &BunqClient,
    account: &MonetaryAccount,
    format: StatementFormat,
    start: &str,
    end: &str,
) -> Result<i64> {
    let statement_request = CreateStatementRequest {
        statement_format: &format.to_string(),
        regional_format: match format {
            StatementFormat::CSV(CsvType::Comma) => "UK_US",
            StatementFormat::CSV(CsvType::Semicolon) => "EUROPEAN",
            _ => "",
        },
        date_start: start,
        date_end: end,
        include_attachment: match format {
            StatementFormat::PDF(v) => v,
            _ => false,
        },
    };

    let body = serde_json::to_string(&statement_request)?;
    let sig = sign(&body, &client.keypair)?;

    let request_url = format!(
        "{}/v1/user/{}/monetary-account/{}/customer-statement",
        BASE_URL,
        client.user_id,
        account.id()
    );

    let mut response = isahc::http::Request::post(request_url)
        .header("X-Bunq-Client-Authentication", &client.token)
        .header("X-Bunq-Client-Signature", &sig)
        .body(body)?
        .send()?;

    if response.status() != StatusCode::OK {
        return Err(anyhow!(response.text()?));
    }

    let response_body = response.text()?;

    //let response_body = r#"{"Response":[{"Id":{"id":4203007}}]}"#;
    //println!("{response_body}");

    let r = deserialize_normal_response::<CreateStatementResponse>(&response_body)?.response;

    let statement_id = r.id.id.id;

    Ok(statement_id)
}

pub(super) fn is_statement_ready(
    client: &BunqClient,
    account: &MonetaryAccount,
    statement_id: i64,
) -> Result<bool> {
    let url = format!(
        "{}/v1/user/{}/monetary-account/{}/customer-statement/{}",
        BASE_URL,
        client.user_id,
        account.id(),
        statement_id
    );

    let mut response = isahc::http::Request::get(url)
        .header("X-Bunq-Client-Authentication", &client.token)
        .body(())?
        .send()?;

    let body = response.text()?;

    let statement = deserialize_normal_response::<Vec<CustomerStatementWrapper>>(&body)?
        .response
        .iter()
        .map(|v| v.customer_statement.clone())
        .take(1)
        .collect::<Vec<CustomerStatement>>()
        .pop()
        .unwrap();

    //let statement = statements.first().unwrap();

    Ok(statement.status == "COMPLETED")
}

pub(super) fn download_statement(
    client: &BunqClient,
    account: &MonetaryAccount,
    id: i64,
) -> Result<String> {
    let content_url = format!(
        "{}/v1/user/{}/monetary-account/{}/customer-statement/{}/content",
        BASE_URL,
        client.user_id,
        account.id(),
        id
    );

    let mut response = isahc::http::Request::get(content_url)
        .header("X-Bunq-Client-Authentication", &client.token)
        .body(())?
        .send()?;

    Ok(response.text()?)
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
pub enum MonetaryAccount {
    MonetaryAccountBank(MonetaryAccountBank),
    MonetaryAccountSavings(MonetaryAccountSavings),
}

impl MonetaryAccount {
    pub fn id(&self) -> i64 {
        match &self {
            MonetaryAccount::MonetaryAccountBank(b) => b.id,
            MonetaryAccount::MonetaryAccountSavings(s) => s.id,
        }
    }
}

#[derive(Deserialize, Debug)]
pub struct MonetaryAccountBank {
    pub id: i64,
    pub description: String,
}

#[derive(Deserialize, Debug)]
pub struct MonetaryAccountSavings {
    pub id: i64,
    pub description: String,
}

pub enum CsvType {
    Comma,
    Semicolon,
}

pub enum StatementFormat {
    CSV(CsvType),
    MT940,
    PDF(bool),
}

impl Display for StatementFormat {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let str = match self {
            StatementFormat::CSV(_) => "CSV",
            StatementFormat::MT940 => "MT940",
            StatementFormat::PDF(_) => "PDF",
        };
        write!(f, "{}", str)
    }
}

impl FromStr for StatementFormat {
    type Err = ();

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_ascii_uppercase().as_str() {
            "CSV" => Ok(StatementFormat::CSV(CsvType::Semicolon)),
            "MT940" => Ok(StatementFormat::MT940),
            "PDF" => Ok(StatementFormat::PDF(false)),
            _ => Err(()),
        }
    }
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
struct CustomerStatementWrapper {
    customer_statement: CustomerStatement,
}

#[derive(Deserialize, Debug, Clone)]
pub struct CustomerStatement {
    pub id: i64,
    status: String,
}

#[derive(Serialize)]
struct CreateStatementRequest<'a> {
    statement_format: &'a str,
    date_start: &'a str,
    date_end: &'a str,
    regional_format: &'a str,
    include_attachment: bool,
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
struct CreateStatementResponse {
    id: Id,
}

//#[derive(Deserialize, Debug)]
#[derive(Deserialize, Debug)]
#[serde(rename_all = "PascalCase")]
struct Id {
    id: Id2,
}

#[derive(Deserialize, Debug)]
struct Id2 {
    id: i64,
}
